// A stream experiment.

#include <cassert>
#include <limits>
#include <memory>
#include <utility>

#define ASSERT_MORE() assert(bool(*this))

namespace stream {

template <typename Str>
class stream_iterator
{
public:
    using value_type = typename Str::value_type;

    stream_iterator() : stream_(nullptr) { }
    explicit stream_iterator(Str& stream) : stream_(&stream) { }

    value_type operator*()
    {
        assert(stream_);
        return (*stream_)();
    }

    stream_iterator& operator++()
    {
        return *this;
    }

    stream_iterator operator++(int)
    {
        return *this;
    }

    bool operator==(const stream_iterator& other)
    {
        return is_empty_() && other.is_empty_();
    }

    bool operator!=(const stream_iterator& other)
    {
        return !(*this == other);
    }

private:
    Str* stream_;

    bool is_empty_() const
    {
        return !(stream_ && bool(*stream_));
    }
};

template <typename Self>
struct stream_base
{
    stream_base() = default;

    stream_base(const stream_base&) = delete;
    stream_base& operator=(const stream_base&) = delete;

    stream_base(stream_base&&) = default;
    stream_base& operator=(stream_base&&) = default;

    stream_iterator<Self> begin()
    {
        return stream_iterator<Self>{static_cast<Self&>(*this)};
    }

    stream_iterator<Self> end()
    {
        return stream_iterator<Self>();
    }
};

template<typename T>
class iota_class : public stream_base<iota_class<T>>
{
public:
    using value_type = T;

    iota_class(T value, T limit)
            : value_(value), limit_(limit)
    { }

    operator bool() const { return value_ < limit_; }
    value_type operator()()
    {
        ASSERT_MORE();
        return value_++;
    }

private:
    T value_;
    T limit_;
};

template<typename T>
iota_class<T> iota(T start = T(), T limit = std::numeric_limits<T>::max())
{
    return {start, limit};
}

template <typename Str>
class take_class : public stream_base<take_class<Str>>
{
public:
    using value_type = typename Str::value_type;

    take_class(size_t size, Str&& base)
            : base_(std::forward<Str>(base)), size_(size)
    { }

    operator bool() { return size_ > 0 && bool(base_); }

    auto operator()() {
        ASSERT_MORE();
        auto&& result = base_();
        --size_;
        return result;
    }

private:
    Str base_;
    size_t size_;
};

template <typename Str>
take_class<Str> take(size_t size, Str&& base)
{
    return {size, std::forward<Str>(base)};
}

template <typename I, typename S = I>
class iterator_pair_class : public stream_base<iterator_pair_class<I, S>>
{
public:
    using value_type = decltype(*std::declval<I>());

    iterator_pair_class(I begin, S end)
            : begin_(std::move(begin)), end_(std::move(end))
    { }

    operator bool() const { return begin_ != end_; }
    value_type operator()()
    {
        ASSERT_MORE();
        return *begin_++;
    }

private:
    I begin_;
    S end_;
};

template <typename I, typename S = I>
iterator_pair_class<I, S> iterator_pair(I begin, S end)
{
    return {std::move(begin), std::move(end)};
};

template <typename C>
iterator_pair_class<typename C::const_iterator>
from_container(const C& container)
{
    using std::begin;
    using std::end;

    return iterator_pair(begin(container), end(container));
}

template <typename Str>
class enumerate_class : public stream_base<enumerate_class<Str>>
{
public:
    using base_type = typename Str::value_type;
    using value_type = std::pair<size_t, base_type>;

    explicit enumerate_class(Str&& base)
            : base_(std::forward<Str>(base)), index_(0)
    { }

    operator bool() const { return bool(base_); }
    value_type operator()()
    {
        ASSERT_MORE();
        auto&& result = base_();
        return {index_++, std::move(result)};
    }

private:
    Str base_;
    size_t index_;
};

template <typename Str>
enumerate_class<Str> enumerate(Str&& base)
{
    return {std::forward<Str>(base)};
}

template <typename Str, typename Fun>
class map_class : public stream_base<map_class<Str, Fun>>
{
public:
    using base_type = typename Str::value_type;
    using value_type = decltype(std::declval<Fun>()(std::declval<base_type>()));

    map_class(Str&& base, Fun&& fun)
            : base_(std::forward<Str>(base)), fun_(std::forward<Fun>(fun))
    { }

    operator bool() const { return bool(base_); }
    value_type operator()()
    {
        ASSERT_MORE();
        return fun_(base_());
    }

private:
    Str base_;
    Fun fun_;
};

template <typename Str, typename Fun>
map_class<Str, Fun> map(Str&& base, Fun&& fun)
{
    return {std::forward<Str>(base), std::forward<Fun>(fun)};
};

template <typename Str, typename Pred>
class filter_class : public stream_base<filter_class<Str, Pred>>
{
public:
    using value_type = typename Str::value_type;

    filter_class(Str&& base, Pred&& pred)
            : base_(std::forward<Str>(base)), pred_(std::forward<Pred>(pred))
    { }

    operator bool() const
    {
        fill_buffer_();
        return buf_ != nullptr;
    }

    value_type operator()()
    {
        fill_buffer_();
        assert(buf_ != nullptr);

        value_type result = *buf_;
        buf_ = nullptr;
        return result;
    }

private:
    mutable Str base_;
    mutable Pred pred_;
    mutable std::unique_ptr<value_type> buf_;

    void fill_buffer_() const
    {
        while (!buf_ && bool(base_)) {
            buf_ = std::make_unique<value_type>(base_());
            if (!pred_(*buf_)) buf_ = nullptr;
        }
    }
};

template<typename Str, typename Pred>
filter_class<Str, Pred> filter(Str&& base, Pred&& pred)
{
    return {std::forward<Str>(base), std::forward<Pred>(pred)};
};

template<typename Str, typename Fun>
void for_each(Str&& stream, Fun fun)
{
    while (stream) {
        fun(stream());
    }
};

template<typename Str, typename OutIter>
void collect(Str&& stream, OutIter out)
{
    for_each(std::forward<Str>(stream), [&](auto&& elem) {
        *out++ = std::move(elem);
    });
};

template<typename Result, typename Str>
Result collect(Str&& stream)
{
    Result result;
    collect(std::forward<Str>(stream), std::back_inserter(result));
    return result;
};

} // end namespace stream
