// Let's try using std::find and std::find_if to write tokenize from HW5:

#include <catch.h>

#include <algorithm>
#include <cctype>
#include <iterator>
#include <string>
#include <vector>

// This is our predicate to pass to std::find_if (below) when searching for
// the end of an unquoted token.
bool is_token_end(char c)
{
    return isspace(c) || c == '"';
}

std::vector<std::string> tokenize(const std::string& input) {
    std::vector<std::string> result;

    auto i = input.begin();

    while (i != input.end()) {
        if (*i == '"') {
            auto right = std::find(i + 1, input.end(), '"');
            if (right == input.end()) throw std::runtime_error("unmatched \"");
            result.emplace_back(i + 1, right);
            i = right + 1;
        } else if (!isspace(*i)) {
            auto right = std::find_if(i + 1, input.end(), is_token_end);
            result.emplace_back(i, right);
            i = right;
        } else ++i;
    }

    return result;
}

/*

Notes on above:

 - input.begin(), std::find, and std::find_if return iterators into the
std::string `input`. These are of type std::string::const_iterator because
`input` is const.

 - std::find takes an open-closed range [i + 1, input.end()) to search in
and returns the iterator to the first occurrence, or input.end() if not
found. Thus, if find returns input.end(), that means the matching
quotation mark wasn't found, and we should throw.

 - std::string has a constructor that takes a closed-open range of
iterators returning characters and constructs a string. Thus, we could write:

    std::string token(i + 1, right);
    result.push_back(token);

Or because result.push_back knows it wants a string, we could shorten this to

    result.push_back({i + 1, right});

This constructs a string, then passes a const string& to push_back, which
copies the string into the vector. Instead, we use the
vector<T>::emplace_back member function, which takes parameters to any
construtor to T and constructs the T directly in the vector. In the string
case, that means that no temporary string is constructed in order to pass
a reference into push_back---instead we just construct one string in place.
It's more efficient and concise.

 */

TEST_CASE("Tokenize")
{
    using V = std::vector<std::string>;

    CHECK(V{} == tokenize(""));
    CHECK(V{} == tokenize("   "));
    CHECK(V{"a"} == tokenize("a"));
    CHECK(V{"abc"} == tokenize("abc"));
    CHECK(V{"a", "bc"} == tokenize("a bc"));
    CHECK(V{"a", "bc"} == tokenize(" a bc"));
    CHECK(V{"a", "bc"} == tokenize("a bc "));
    CHECK(V{"a", "bc"} == tokenize(" a bc "));

    CHECK_THROWS_AS(tokenize("\""), std::runtime_error);
    CHECK(V{""} == tokenize("\"\""));
    CHECK(V{" "} == tokenize("\" \""));
    CHECK(V{" "} == tokenize(" \" \""));
    CHECK(V{" "} == tokenize("\" \" "));
    CHECK(V{" "} == tokenize("   \" \" "));
    CHECK(V{" ", "a b"} == tokenize("   \" \"\"a b\" "));
    CHECK(V{" ", "a b"} == tokenize("   \" \"  \"a b\" "));

    CHECK(V{"a b", "c"} == tokenize("\"a b\"c"));
    CHECK(V{"a", "b c"} == tokenize("a\"b c\""));
}

// We can write a more generic version of tokenize.

template <typename FwdIter, typename OutIter>
void tokenize(FwdIter i, FwdIter limit, OutIter out) {
    while (i != limit) {
        if (*i == '"') {
            FwdIter right = std::find(++i, limit, '"');
            if (right == limit) throw std::runtime_error("unmatched \"");
            *out++ = {i, right};
            i = ++right;
        } else if (!isspace(*i)) {
            FwdIter right = std::find_if(i + 1, limit, is_token_end);
            *out++ = {i, right};
            i = right;
        } else ++i;
    }
}

TEST_CASE("Generic_tokenize")
{
    using V = std::vector<std::string>;

    V out;
    const std::string str = "This can be \"parsed\" into tokens.";
    tokenize(str.begin(), str.end(), std::back_inserter(out));

    V expected{"This", "can", "be", "parsed", "into", "tokens."};
    CHECK(expected == out);
}
